﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

namespace HeuristicLab.Persistence.Default.Xml {

  public class AsyncBuffer<T> : IDisposable, IEnumerable<T>, IEnumerable {

    private IEnumerator<T> it;

    private Queue<T> queue;
    private object queueLock;
    private AutoResetEvent elementAdded;

    private Thread thread;
    private ManualResetEventSlim done;

    private Exception exception;

    private class LockIndicator {
      public bool HasLock;
      public LockIndicator() { HasLock = false; }
    }

    public AsyncBuffer(IEnumerable<T> source) {
      it = source.GetEnumerator();
      queue = new Queue<T>();
      queueLock = new object();
      elementAdded = new AutoResetEvent(false);
      thread = new Thread(Run);
      done = new ManualResetEventSlim(false);
      thread.Start();
    }

    private void Run() {
      try {
        while (it.MoveNext()) {
          T value = it.Current;
          lock (queueLock) {
            queue.Enqueue(value);
            elementAdded.Set();
          }
        }
      } catch (Exception x) {
        exception = x;
      }
      done.Set();
    }

    #region IEnumerable<T> Members
    public IEnumerator<T> GetEnumerator() {
      LockIndicator li = new LockIndicator();
      try {
        return RealGetEnumerator(li);
      } catch {
        if (li.HasLock)
          Monitor.Exit(queueLock);
        throw;
      }
    }

    private IEnumerator<T> RealGetEnumerator(LockIndicator li) {
      Monitor.Enter(queueLock, ref li.HasLock);
      while (!done.IsSet || queue.Any()) {
        while (!queue.Any() && !done.IsSet) {
          li.HasLock = false;
          Monitor.Exit(queueLock);
          AutoResetEvent.WaitAny(new WaitHandle[] { done.WaitHandle, elementAdded });
          Monitor.Enter(queueLock, ref li.HasLock);
        }
        if (queue.Any()) {
          T element = queue.Peek();
          li.HasLock = false;
          Monitor.Exit(queueLock);
          yield return element;
          Monitor.Enter(queueLock, ref li.HasLock);
          queue.Dequeue();
        }
      }
      li.HasLock = false;
      Monitor.Exit(queueLock);
      if (exception != null)
        throw exception;
    }
    #endregion

    #region IEnumerable Members
    IEnumerator IEnumerable.GetEnumerator() {
      return GetEnumerator();
    }
    #endregion

    #region IDisposable Members
    public void Dispose() {
      if (elementAdded != null)
        elementAdded.Dispose();
      if (done != null)
        done.Dispose();
    }
    #endregion
  }
}
